#
# Ffidl interface to Tcl8.2
#
# Run time support for Ffidl.
#
package provide Ffidlrt 0.1
package require Ffidl 0.1

namespace eval ::ffidlrt:: {
    #
    # found libraries
    # this array is used by the ffidl-find-lib
    # abstraction to store the resolved lib paths
    #
    # currently wired for my linux box
    # CHANGE - put your resolved lib paths here
    #
    array set libs {
	c /lib/libc.so.6
	m /lib/libm.so.6
	tcl /usr/local/lib/libtcl8.2.so
	tk /usr/local/lib/libtk8.2.so
	gdbm /usr/lib/libgdbm.so
	gmp /usr/lib/libgmp.so.2
	ffidl ../libFfidl0.1.so
	ffidl_test ../libFfidl_test0.1.so
    }
    #
    # found types
    # these arrays are used by the ffidl-find-type
    # abstraction to store resolved system types
    # and whether they have already been defined
    # with ffidl-typedef
    #
    # currently wired for my linux box
    # CHANGE - put your resolved system types here
    #
    array set types {
	size_t int
	clock_t long
	time_t long
	timeval {time_t time_t}
    }
    array set typedefs {
    }
}

#
# find a shared library given a root name
# this is an abstraction in search of a
# solution.
#
# currently wired for my linux box
#
proc ffidl-find-lib {root} {
    upvar \#0 ::ffidlrt::libs libs
    if { ! [info exists libs($root)]} {
	error "ffidl-find-lib $root - no mapping defined for $root"
    }
    set libs($root)
}

#
# find a typedef for a standard type
# and define it with ffidl-typedef
# if not already done
#
# currently wired for my linux box
#
proc ffidl-find-type {type} {
    upvar \#0 ::ffidlrt::types types
    upvar \#0 ::ffidlrt::typedefs typedefs
    if { ! [info exists types($type)]} {
	error "ffidl-find-type $type - no mapping defined for $type"
    }
    if { ! [info exists typedefs($type)]} {
	eval ffidl-typedef $type $types($type)
	set typedefs($type) 1
    }
}
    
#
# get the address of the string rep of a Tcl_Obj
# get the address of the unicode rep of a Tcl_Obj
# get the address of the bytearray rep of a Tcl_Obj
#
# CAUTION - anything which alters the Tcl_Obj may
# invalidate the results of this function.  Use
# only in circumstances where the Tcl_Obj will not
# be modified in any way.
#
# CAUTION - the memory pointed to by the addresses
# returned by ffidl-get-string and ffidl-get-unicode
# is managed by Tcl, the contents should never be
# modified.
#
# The memory pointed to by ffidl-get-bytearray may
# be modified if care is taken to respect its size,
# and if shared references to the bytearray object
# are known to be compatible with the modification.
#

ffidl-proc ffidl-get-string {pointer-obj} pointer [ffidl-symbol [ffidl-find-lib tcl] Tcl_GetString]
ffidl-proc ffidl-get-unicode {pointer-obj} pointer [ffidl-symbol [ffidl-find-lib tcl] Tcl_GetUnicode]
ffidl-proc ffidl-get-bytearray-from-obj {pointer-obj pointer-var} pointer [ffidl-symbol [ffidl-find-lib tcl] Tcl_GetByteArrayFromObj]

proc ffidl-get-bytearray {obj} {
    set len [binary format [ffidl-info format int] 0]
    ffidl-get-bytearray-from-obj $obj len
}

#
# create a new string Tcl_Obj
# create a new unicode Tcl_Obj
# create a new bytearray Tcl_Obj
#
# I'm not sure if these are actually useful
#

ffidl-proc ffidl-new-string {pointer int} pointer-obj [ffidl-symbol [ffidl-find-lib tcl] Tcl_NewStringObj]
ffidl-proc ffidl-new-unicode {pointer int} pointer-obj [ffidl-symbol [ffidl-find-lib tcl] Tcl_NewUnicodeObj]
ffidl-proc ffidl-new-bytearray {pointer int} pointer-obj [ffidl-symbol [ffidl-find-lib tcl] Tcl_NewByteArrayObj]

#
# access the standard allocator, malloc, free, realloc
#
ffidl-find-type size_t
ffidl-proc ffidl-malloc {size_t} pointer [ffidl-symbol [ffidl-find-lib c] malloc]
ffidl-proc ffidl-realloc {pointer size_t} pointer [ffidl-symbol [ffidl-find-lib c] realloc]
ffidl-proc ffidl-free {pointer} void [ffidl-symbol [ffidl-find-lib c] free]

#
# Copy some memory at some location into a Tcl bytearray.
#
# Needless to say, this can be very hazardous to your
# program's health if things aren't sized correctly.
#

ffidl-proc ffidl-memcpy {pointer-var pointer int} pointer [ffidl-symbol [ffidl-find-lib c] memcpy]

proc ffidl-peek {address nbytes} {
    set dst [binary format x$nbytes]
    ffidl-memcpy dst $address $nbytes
    set dst
}

#
# convert raw pointers, as integers, into Tcl_Obj's
#
ffidl-proc ffidl-pointer-into-string {pointer} pointer-utf8 [ffidl-symbol [ffidl-find-lib ffidl] ffidl_pointer_pun]
ffidl-proc ffidl-pointer-into-unicode {pointer} pointer-utf16 [ffidl-symbol [ffidl-find-lib ffidl] ffidl_pointer_pun]
proc ffidl-pointer-into-bytearray {pointer length} {
    set bytes [binary format x$length]
    ffidl-memcpy [ffidl-get-bytearray $bytes] $pointer $length
    set bytes
}
